#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Young King'
__email__ = 'y@zhihu.com'
__version__ = '0.1.0'

from .lazyconn import LazyConnection, thread_safe

__all__ = ['LazyConnection', 'thread_safe']
